#!/usr/bin/env python

# Send a bug to the kanban board
#
# Usage:
# ./bug2kb.py <bugnum>

import os.path
import sys
from os.path import dirname

from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import (Credentials, Consumer, AccessToken)
from lazr.restfulclient.errors import HTTPError

from config import load_app_config
from leankit import LeanKitBase


class UnknownUserException(Exception):
    pass


class BugAlreadyOnBoardException(Exception):
    pass


class BugToBoard(LeanKitBase):
    _identifiers = None

    def create_task_for_bug(self, bug):
        if self.bug_already_on_board(bug):
            raise BugAlreadyOnBoardException("Bug %s is already on the " \
                                             "board" % bug.id)
        if self.config.values('__main__')['debug']:
            print "Creating task for bug #%s" % bug.id
        data = self.map_bug_to_task(bug)
        path = '/Kanban/Api/Board/%s/AddCard/Lane/%s/Position/%s' % (
            self.config.values('board')['id'],
            self.config.values('board')['default_lane_id'], 0)
        task = self.do_request(path, data)
        return None

    def bug_already_on_board(self, bug):
        def card_in_lane(bug, card, lane):
            bug_url = self.config.values('lp')['bug_url'] % bug.id
            if card['ExternalSystemUrl'] == bug_url:
                return True

        lanes = []
        lanes.extend(self.board['ReplyData'][0]['Lanes'])
        lanes.extend(self.board['ReplyData'][0]['Backlog'])
        lanes.extend(self.board['ReplyData'][0]['Archive'])

        for lane in lanes:
            for card in lane['Cards']:
                if card_in_lane(bug, card, lane):
                    return True

        return False

    def map_bug_to_task(self, bug):
        return {'Id': 0,
                'Title': self.get_title_from_bug(bug),
                'Description': bug.description,
                'TypeId': self.get_type_id_from_bug(bug),
                'Size': self.get_sp_for_bug(bug),
                'Priority': self.get_priority_id_from_bug(bug),
                'AssignedUserId': self.get_kb_userid_from_bug(bug),
                'IsBlocked': False,
                'ExternalSystemUrl': self.config.values('lp')['bug_url'] % bug.id}

    def get_title_from_bug(self, bug):
        mappings = self.config.values('mappings')['titles']
        project_name = bug.bug_tasks[0].bug_target_name
        project_tag = mappings.get(project_name, '')
        if project_tag:
            project_tag = "[{0}]".format(project_tag)
        title = "{0}[{1}] {2}".format(
            project_tag, bug.id, bug.title)
        return title

    def get_type_id_from_bug(self, bug):
        card_types = self.identifiers['ReplyData'][0]['CardTypes']
        default_type = self.config.values('board')['default_type']
        mappings = self.config.values('mappings')['types']
        for tag in mappings:
            if tag in bug.tags:
                for card_type in card_types:
                    if card_type['Name'] == mappings[tag]:
                        return card_type['Id']
        for card_type in card_types:
            if card_type['Name'] == default_type:
                return card_type['Id']
        return None

    def get_sp_for_bug(self, bug):
        for tag in bug.tags:
            if tag.startswith('sp-'):
                try:
                    return int(tag[3:])
                except:
                    next
        return int(self.config.values('lp')['default_bug_sp'])

    def get_kb_userid_from_bug(self, bug, default_id=0):
        name = '<unassigned>'
        for task in bug.bug_tasks:
            if task.bug_target_name == self.config.values('board')['projects']:
                if self.debug:
                    print "Target: %s" % task.bug_target_name
                    print "Projs: %s" % self.config.values('board')['projects']
                try:
                    name = task.assignee.name
                except AttributeError:
                    pass
                if self.debug:
                    print "Assignee: %s" % name
        try:
            email = self.config.values('mappings')['users'][name]
            if self.debug:
                print "Found email %s for user" % email
            for user in self.identifiers['ReplyData'][0]['BoardUsers']:
                if user['Name'] == email:
                    if self.debug:
                        print "Found id %s for user %s" % (user['Id'], name)
                    return user['Id']
            if self.debug:
                print "Couldn't find user"
            raise UnknownUserException()
        except (KeyError, UnknownUserException):
            if self.debug:
                print "Failed to find id for user %s" % name
            return default_id

    def get_priority_id_from_bug(self, bug):
        importance = self.get_importance_from_bug(bug)
        return self.config.values('mappings')['priorities'][importance]

    def get_importance_from_bug(self, bug):
        for task in bug.bug_tasks:
            if task.bug_target_name in self.config.values('board')['projects']:
                return task.importance
        return 'Undecided'


def tag_to_bugs(lp, proj_tag):
    project, tag = proj_tag.split(':')
    lp_project = lp.projects[project]
    bugs = lp_project.searchTasks(tags=tag)
    return (b.bug for b in bugs)


def main():
    bugs = set([])
    tags = set([])
    for arg in sys.argv[1:]:
        if arg.startswith('tag:'):
            proj_tag = arg.split(':', 1)[1]
            tags.add(proj_tag)
        else:
            try:
                bugs.add(int(arg))
            except (IndexError, ValueError):
                pass

    if len(bugs) < 1 and len(tags) < 1:
        print "Usage: bug2kb.py [<bugnumber>, ...] [options]"
        sys.exit()

    config = load_app_config()
    lp_config = config.values('lp')

    try:
        c = Credentials()
        c.consumer = Consumer(lp_config['app_name'], '')
        c.access_token = AccessToken(lp_config['access_token'],
                                     lp_config['access_secret'])
        lp = Launchpad(c, service_root=lp_config['service_root'],
                       cache=lp_config['cache_dir'])
    except HTTPError, e:
        if config.values('__main__')['debug']:
            raise e
        else:
            print "Failed to connect to Launchpad (%s).\nEnable debug for " \
                  "more verbose error reporting." % lp_config['service_root']
        sys.exit()

    # map tags to bugs
    for tag in tags:
        tagged_bugs = tag_to_bugs(lp, tag)
        bugs.update(set(tagged_bugs))

    b2b = BugToBoard(config, lp)
    for bugid in bugs:
        try:
            b2b.create_task_for_bug(lp.bugs[bugid])
        except BugAlreadyOnBoardException, e:
            print "Error: %s" % e


if __name__ == '__main__':
    main()
