#!/usr/bin/env python

import sys
from os.path import dirname

from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import (Credentials, Consumer, AccessToken)
from lazr.restfulclient.errors import HTTPError

from config import load_app_config
from leankit import LeanKitBase


class SyncBoardToLP(LeanKitBase):

    def board_to_lp(self):
        for card in self.all_cards():
            if card.is_bug_card(self.config):
                self.update_lp_bug_from_card(card)

    def update_lp_bug_from_card(self, card):
        bug_url = self.config.values('lp')['bug_url'] % ''
        bug_id = card.external_system_url[len(bug_url):]
        bug = self.lp.bugs[bug_id]
        changed = False

        projects = self.config.values('board')['projects']
        # This next set of actions is only conducted on the main target project.
        # We also need to loop throught the related_tasks collection to make the
        # same changes on each project we care about in those too.
        if hasattr(bug, 'target') and bug.target.name in projects:
            self.set_bug_task_info(bug_id, bug, card)

        for i in range(0, len(bug.bug_tasks)):
            task = bug.bug_tasks[i]
            if task.target.name in self.config.values('board')['projects']:
                self.set_bug_task_info(bug_id, task, card)

        # Set tag based on type
        tag = self.get_tag_for_type_id(card.type_name)
        if not tag in bug.tags:
            bug_tags = bug.tags
            for t in self.config.values('mappings')['types']:
                if t == tag:
                    bug_tags.append(t)
                elif t in bug_tags:
                    bug_tags.remove(t)
            bug.tags = bug_tags
            if self.debug:
                print "Setting tags for %s to %s" % (bug_id, bug_tags)
            changed = True

        # Set story points
        points = self.get_story_points_from_card(card)
        tag = 'sp-%s' % points
        if not tag in bug.tags:
            bug_tags = bug.tags
            for t in bug_tags:
                if t.startswith('sp-'):
                    bug_tags.remove(t)
            bug_tags.append(tag)
            bug.tags = bug_tags
            if self.debug:
                print "Setting tags for %s to %s" % (bug_id, bug_tags)
            changed = True

        if changed:
            if self.debug:
                print "Saving bug %s" % bug_id
            bug.lp_save()

    def set_bug_task_info(self, bug_id, bug, card):
        # Set importance
        importance = self.get_importance_from_priority(card.priority)
        if importance != bug.importance and importance is not None:
            if self.debug:
                print "Setting bug %s to importance %s for task %s" % (bug_id, importance, bug.target.name)
            bug.importance = importance
            changed = True

        # Set status
        status = self.get_status_for_lane(card.lane_id)
        if status != bug.status and status is not None:
            if self.debug:
                print "Setting bug %s to status %s for task %s" % (bug_id, status, bug.target.name)
            bug.status = status
            changed = True

        # Set assignee
        assignee = self.get_assignee_from_card(card.assigned_user_id)
        if bug.assignee is None and assignee or bug.assignee is not None and \
                assignee != bug.assignee.name and assignee is not None:
            if self.debug:
                print "Assigning bug %s to %s for task %s" % (bug_id, assignee, bug.target.name)
            bug.assignee = self.lp.people[assignee]
            changed = True

        bug.lp_save()

    def get_tag_for_type_id(self, type_id):
        types = self.config.values('mappings')['types']
        for tag in types:
            if types[tag] == type_id:
                return tag
        return None

    def get_importance_from_priority(self, priority):
        priorities = self.config.values('mappings')['priorities']
        #priorities.reverse()
        for p in priorities:
            if priority == priorities[p]:
                return p
        return None

    def get_status_for_lane(self, lane_id):
        lane_id = str(lane_id)
        lanes = self.config.values('mappings')['lanes']
        for l in lanes:
            if l == lane_id:
                return lanes[l]
        return None

    def get_assignee_from_card(self, user_id):
        email = self.get_email_for_user_id(user_id)
        users = self.config.values('mappings')['users']
        for u in users:
            if users[u] == email:
                return u
        return None

    def get_story_points_from_card(self, card):
        return card.size or 1

    def get_email_for_user_id(self, user_id):
        users = self.identifiers['ReplyData'][0]['BoardUsers']
        for u in users:
            if u['Id'] == user_id:
                return u['Name']
        return None


def main():
    config = load_app_config()
    lp_config = config.values('lp')

    try:
        c = Credentials()
        c.consumer = Consumer(lp_config['app_name'], '')
        c.access_token = AccessToken(lp_config['access_token'],
                                     lp_config['access_secret'])
        lp = Launchpad(c, service_root=lp_config['service_root'],
                       cache=lp_config['cache_dir'])
    except HTTPError, e:
        if config.values('__main__')['debug']:
            raise e
        else:
            print "Failed to connect to Launchpad (%s).\nEnable debug for " \
                  "more verbose error reporting." % lp_config['service_root']
        sys.exit()

    sync = SyncBoardToLP(config, lp)
    sync.board_to_lp()


if __name__ == '__main__':
    main()
